# -*- coding: utf-8 -*-
import collections
import math
import warnings
import json

"""Main module."""


def sliding(string, n):
    for i in range(len(string) - n + 1):
        yield string[i : i + n]


class Model(object):
    """A counter for ngrams"""

    def __init__(self, size=2.0, name=None):
        self.size = size
        self.default_count = 0.5  # make a parameter?
        self.counter = collections.Counter()
        self.total = 0
        self.name = name

    def __repr__(self):
        return "<Model {}>".format(self.name)

    def update(self, string, muliplier=1):
        """
        Public: update a counter with a string, and a multiplier

        Examples

        counter = NGramCounter(2)
        counter.update_with_multiplier('01234', 1)

        string: The String to update with
        multiplier: The Integer describing how much weight (will often be 1)
        """
        for substr in sliding(string, self.size):
            self.counter[substr] += muliplier
            self.total += muliplier

    def count(self, ngram):
        """
        Public: get count for string, with default

         Examples

         counter = NGramCounter.new(2)
         counter.update('01234')
         counter.count('01')
         #=> 1
         counter.count('bob')
         #=> 0.5

         ngram: The String to check

        """
        return self.counter.get(ngram, self.default_count)

    def log_prob(self, key):
        if self.total == 0 or not key:
            return -math.inf
        count = self.count(key)
        return math.log(count, 2.0) - math.log(self.total, 2.0)

    def dump(self, io):
        for key, value in self.counter.items():
            io.write("{}\t{}\t{}\n".format(self.size, key, value))

    def dump_json(self, io):
        for key, value in self.counter.items():
            io.write(
                json.dumps(
                    {
                        "model_name": self.name,
                        "ngram_size": self.size,
                        "ngram": key,
                        "count": value,
                    }
                )
                + "\n"
            )

    def predict(self, string):
        n_ngrams = 0
        log_prob_total = 0.0
        for ngram in sliding(string, self.size):
            n_ngrams += 1
            log_prob_total += self.log_prob(ngram)
        log_prob_average = log_prob_total / n_ngrams
        return {
            "log_prob_total": log_prob_total,
            "log_prob_average": log_prob_average,
            "size": int(n_ngrams),
        }

    def entropy(self, string):
        return -self.predict(string)["log_prob_average"]

    def read(self, io):
        n = 0
        for line in io:
            n += 1
            try:
                _, ngram, count = line.strip().split("\t")
                count = float(count)
                self.counter[ngram] += count
                self.total += count
            except ValueError as err:
                warnings.warn("Invalid line at {}. Error: {}".format(n, err))

    def read_json(self, io, count_key="count", ngram_key="ngram"):
        n = 0
        for line in io:
            n += 1
            data = json.loads(line.strip())
            count = data[count_key]
            ngram = data[ngram_key]
            self.counter[ngram] += count
            self.total += count

    def read_from_string(self, string):
        n = 0
        for line in string.split("\n"):
            n += 1
            _, ngram, count = line.strip().split("\t")
            count = float(count)
            self.counter[ngram] += count
            self.total += count

    def train(self, io):
        for line in io:
            self.update(line.strip())

    def train_with_multiplier(self, io):
        n = 0
        for line in io:
            n += 1
            try:
                text, count = line.strip().split("\t")
                count = float(count)
                self.update(text, count)
            except ValueError as err:
                warnings.warn("Invalid line at {}. Error: {}".format(n, err))


def read_models_from_json(
    io,
    count_key="count",
    ngram_key="ngram",
    model_name_key="model_name",
    ngram_size_key="ngram_size",
    ngram_size=None,
):
    models = {}
    for line in io:
        data = json.load(line.strip())
        ngram = data[ngram_key]
        count = data[count_key]
        model_name = data[model_name_key]
        ngram_size = data[ngram_size_key]
        if model_name not in models:
            models[model_name] = Model(size=ngram_size, name=model_name)
        model = models[model_name]
        model.update(ngram, count)
    return models



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pytropic` package."""


import tempfile
import unittest
import os
from pytropic import pytropic


class TestPytropic(unittest.TestCase):
    """Tests for `pytropic` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_sliding(self):
        assert list(pytropic.sliding("test", 2)) == ["te", "es", "st"]

    def test_read(self):
        m = pytropic.Model(2)
        cwd = os.path.dirname(os.path.realpath(__file__))
        p = os.path.join(cwd, "data", "google_books_2.tsv")
        with open(p) as f:
            m.read(f)
        assert m.entropy("fish") > 0

    def test_read_corpus(self):
        en = pytropic.Model(2, name="English sample")
        fr = pytropic.Model(2, name="French sample")
        cwd = os.path.dirname(os.path.realpath(__file__))
        en_path = os.path.join(cwd, "data", "bible-english-sample.txt")
        fr_path = os.path.join(cwd, "data", "bible-french-sample.txt")
        with open(en_path) as f:
            en.train(f)
        with open(fr_path) as f:
            fr.train(f)
        assert en.entropy("fish") < fr.entropy("fish")
        assert en.entropy("poisson") > fr.entropy("poisson")

    def test_read_write_json(self):
        fr = pytropic.Model(2, name="French sample")
        cwd = os.path.dirname(os.path.realpath(__file__))
        fr_path = os.path.join(cwd, "data", "bible-french-sample.txt")
        with open(fr_path) as f:
            fr.train(f)
        fr2 = pytropic.Model(2, name="Read French sample")
        tf = tempfile.mkstemp(text=True)[1]
        with open(tf, "w") as f:
            fr.dump_json(f)
        with open(tf) as f:
            fr2.read_json(f)
        assert fr.entropy("poisson") == fr2.entropy("poisson")

    def test_read_json(self):
        fr = pytropic.Model(2, name="French sample")
        cwd = os.path.dirname(os.path.realpath(__file__))
        fr_path = os.path.join(cwd, "data", "bible-french-sample.json")
        with open(fr_path) as f:
            fr.read_json(f)
        assert fr.entropy("poisson") < fr.entropy("fish")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []

setup_requirements = []

test_requirements = ["pytest"]

setup(
    name="pytropic",
    version="1.0.0",
    description="Train and predict string entropy based on character n-grams",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    author="Will Fitzgerald",
    author_email="will.fitzgerald@gmail.com",
    url="https://github.com/willf/pytropic",
    packages=find_packages(include=["pytropic"]),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="pytropic",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)
